/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#if defined(__PYTHON__)
  #include <Python.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "NCanUtils.h"
#include "KElsterTable.h"

#include "ElsterTable.inc"

static signed short ElsterTabIndex[0x10000];

static const char * ElsterTypeStr[] =
{
  "et_default",
  "et_dec_val",
  "et_cent_val",
  "et_mil_val",
  "et_byte",
  "et_double_val",
  "et_triple_val",
  "et_little_endian",
  "et_zeit",
  "et_datum",
  "et_time_domain",
  "et_dev_nr",
  "et_err_nr"
};

bool SetElsterTabIndex()
{
  for (int i = 0; i <= (int) High(ElsterTabIndex); i++)
    ElsterTabIndex[i] = -1;
    
  for (int i = 0; i <= (int) High(ElsterTable); i++)
    if (ElsterTabIndex[ElsterTable[i].Index] == -1)
      ElsterTabIndex[ElsterTable[i].Index] = i;
    
  return true;
}

static bool ElsterTabIndexOk = SetElsterTabIndex();

void SetValueType(char * Val, unsigned char Type, unsigned short Value)
{
   if (Value == 0x8000)
     strcpy(Val, "not available");
   else
   switch (Type)
   {
     case et_byte:
       sprintf(Val, "%d", (signed char)Value);
       break;
       
     case et_dec_val:
       sprintf(Val, "%.1f", ((double)((signed short)Value)) / 10.0);
       break;
              
     case et_cent_val:
       sprintf(Val, "%.2f", ((double)((signed short)Value)) / 100.0);
       break;

     case et_mil_val:
       sprintf(Val, "%.3f", ((double)((signed short)Value)) / 1000.0);
       break;

     case et_little_endian:
       sprintf(Val, "%d", (Value >> 8) + 256*(Value & 0xff));
       break;
            
     case et_zeit:
       sprintf(Val, "%2.2d:%2.2d", Value & 0xff, Value >> 8);
       break;

     case et_datum:
       sprintf(Val, "%2.2d.%2.2d.", Value >> 8, Value & 0xff);
       break;

     case et_time_domain:
       if (Value & 0x8080)
         strcpy(Val, "not used time domain");
       else
         sprintf(Val, "%2.2d:%2.2d-%2.2d:%2.2d",
                 (Value >> 8) / 4, 15*((Value >> 8) % 4),
                 (Value & 0xff) / 4, 15*(Value % 4));
       break;
       
     case et_dev_nr:
       if (Value >= 0x80)
         strcpy(Val, "--");
       else
         sprintf(Val, "%d", Value + 1);
       break;
       
     case et_err_nr:
     {
       int idx = -1;
       for (unsigned i = 0; i <= High(ErrorList); i++)
         if (ErrorList[i].Index == Value)
         {
           idx = i;
           break;
         }
       if (idx >= 0)
         strcpy(Val, ErrorList[idx].Name);
       else
         sprintf(Val, "ERR %d", Value);
       break;
     }

     case et_default:
     default:
       sprintf(Val, "%d", (signed short)Value);
       break;
   }
}

int GetElsterTableIndex(unsigned short Index)
{
  return ElsterTabIndex[Index];
}

const char * GetElsterTableName(unsigned short Index)
{
  int ind = GetElsterTableIndex(Index);
  if (ind >= 0)
    return ElsterTable[ind].Name;
  else
    return "";
}

ElsterType GetElsterType(const char * str)
{
  if (str)
  {
    for (unsigned i = 0; i <= High(ElsterTypeStr); i++)
      if (!strcmp(ElsterTypeStr[i], str))
        return (ElsterType) i;
  }
  return et_default;
}

const char * ElsterTypeToName(unsigned Type)
{
  if (Type <= High(ElsterTypeStr))
    return ElsterTypeStr[Type];
  
  return ElsterTypeStr[et_default];
}

const ElsterIndex * GetElsterIndex(unsigned short Index)
{
  int Ind = GetElsterTableIndex(Index);
  
  return Ind >= 0 ? &ElsterTable[Ind] : NULL;
}

bool FormElsterTable(const KCanFrame & Frame, char * str)
{
  *str = 0;
  
  if (Frame.Len != 7)
    return false;
  
  int Index = Frame.Index();
  if (Index < 0)
    return false;

  int ind = GetElsterTableIndex((unsigned short)Index);
  if (ind < 0)
    return false;
  
  char val[64];
  
  SetValueType(val, ElsterTable[ind].Type, Frame.Value());
  sprintf(str, "%s = %s", ElsterTable[ind].Name, val);
  
  return true;
}


