/*
 *
 *  Copyright (C) 2014 Jrg Mller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "NCanUtils.h"
#include "KScanTable.h"

#include "scan_data.inc"

static struct
{
  int start_index;
  int length;
} scan_table[256]; // use ShortCanId

static scan_data * _scan_data_table = scan_data_table;
static int scan_tab_len = High(scan_data_table) + 1;
static bool IsInitalized = Init_scan_table();

bool Init_scan_table()
{
  for (int i = 0; i <= (int) High(scan_table); i++)
  {
    scan_table[i].start_index = -1;
    scan_table[i].length = 0;
  }
  for (int i = 0; i < scan_tab_len; i++)
  {
    int id = ShortCanId(_scan_data_table[i].id);
    if (scan_table[id].start_index == -1)
    {
      scan_table[id].start_index = i;
    }
    scan_table[id].length++;
  }
  IsInitalized = true;
  return true;
}

bool Get_scan_value(const KComfortFrame & send, KComfortFrame & recv)
{
  KCanFrame SendFrame;
  KCanFrame RecvFrame;

  if (!IsInitalized)
    return false;

  bool Ok = send.SetCanFrame(SendFrame);
  KComfortFrame::msg_type type = send.GetMsgType();

  unsigned char id = ShortCanId(SendFrame.ReceiverId());

  int table_ind = scan_table[id].start_index;
  if (!Ok || table_ind == -1 || type == KComfortFrame::mt_invalid)
    return false;

  unsigned short ind = SendFrame.Index();
  RecvFrame = SendFrame;
  RecvFrame.Id = SendFrame.ReceiverId();
  RecvFrame.Data[0] = 0xd2;

  int data_index = -1;
  for (int i = 0; i < scan_table[id].length; i++)
    if (_scan_data_table[table_ind + i].index == ind)
    {
      data_index = table_ind + i;
      break;
    }

  unsigned short val = 0x8000;
  if (data_index != -1)
    val = _scan_data_table[data_index].value;

  if (type == KComfortFrame::mt_get)
  {
    RecvFrame.SetValue(val);
    return recv.GetCanFrame(RecvFrame);
  }
  if (type == KComfortFrame::mt_change ||
      type == KComfortFrame::mt_short_change)
  {
    if (val != 0x8000)
    {
      _scan_data_table[data_index].value = SendFrame.Value();
      recv.SetOk(true);
    } else
      recv.SetOk(false);
    return true;
  }
  return false;
}

bool LoadScanTable(const char * Filename)
{
  if (_scan_data_table &&
      _scan_data_table != scan_data_table)
    delete [] _scan_data_table;
  
  _scan_data_table = NULL;
  IsInitalized = false;
  scan_tab_len = 0;
  
  FILE * fin = fopen(Filename, "rb");
  if (!fin)
    return false;
  
  long flen = 0;
  if (!fseek(fin, 0, SEEK_END))
  {
    flen = ftell(fin);
    if (fseek(fin, 0, SEEK_SET))
      flen = 0;
  }
  if (flen <= 0)
  {
    fclose(fin);
    return false;
  }
  
  unsigned char * buffer = new unsigned char [flen+1];
  buffer[flen] = 0;
  long pos = 0;
  while (pos < flen)
  {
    
    long l = fread(buffer + pos, 1, flen > 10000 ? 10000 : flen, fin);
    if (l <= 0)
      break;
    pos += l;
  }
  
  bool Ok = flen == pos;
  int count = 0;
  if (Ok)
  {
    pos = 0;
    while (pos < flen)
    {
      count++;
      while (buffer[pos] != '\r' && buffer[pos] != '\n')
        pos++;
      if (buffer[pos] == '\r' || buffer[pos] == '\n')
        pos++;
    }

    _scan_data_table = new scan_data[count];
    pos = 0;
    while (pos < flen)
    {
      scan_data newtab;
      
      while (buffer[pos] != '\r' && buffer[pos] != '\n')
        pos++;
      if (buffer[pos] == '\r' || buffer[pos] == '\n')
        pos++;
      while (buffer[pos] == ' ') pos++;
      
      bool IsLine = buffer[pos] == '{';
      if (IsLine) pos++;
      while (buffer[pos] == ' ') pos++;
      
      int res;
      unsigned inc = NCanUtils::GetInt(res, (char *)(buffer + pos));
      if (!inc)
        IsLine = false;
      newtab.id = (unsigned short) res;
      pos += inc;
      while (buffer[pos] == ' ' || buffer[pos] == ',') pos++;
      
      inc = NCanUtils::GetInt(res, (char *)(buffer + pos));
      if (!inc)
        IsLine = false;
      newtab.index = (unsigned short) res;
      pos += inc;
      while (buffer[pos] == ' ' || buffer[pos] == ',') pos++;

      inc = NCanUtils::GetInt(res, (char *)(buffer + pos));
      if (!inc)
        IsLine = false;
      newtab.value = (unsigned short) res;
      pos += inc;
      
      if (IsLine)
      {
        _scan_data_table[scan_tab_len++] = newtab;
      }
    }
  }
  
  delete [] buffer;
  fclose(fin);
  
  if (Ok)
    Ok = Init_scan_table();
  
  return Ok;
}

static bool write(const char * str, FILE * fout)
{
  unsigned long len = strlen(str);
  return fwrite(str, 1, len, fout) == len;
}

bool SaveScanTable(const char * Filename)
{
  if (!_scan_data_table || !IsInitalized || scan_tab_len <= 0)
    return false;

  FILE * fout = fopen(Filename, "wb");
  if (!fout)
    return false;

  unsigned short last = 0;
  // static scan_data scan_data_table[] =
  // {
  //   { 0x000, 0x0a26, 0x0001},
  bool Ok = write("static scan_data scan_data_table[] =\n"
                  "{\n", fout);
  for (int i = 0; Ok && i < scan_tab_len; i++)
  {
    char str[64];
    if (last != _scan_data_table[i].id && i > 0)
      write("\n", fout);
    sprintf(str, "  { 0x%3.3x, 0x%4.4x, 0x%4.4x}%c\n",
            _scan_data_table[i].id,
            _scan_data_table[i].index,
            _scan_data_table[i].value,
            i != scan_tab_len-1 ? ',' : ' ');
    Ok = write(str, fout);
    last = _scan_data_table[i].id;
  }
  if (Ok)
    Ok = write("};\n", fout);

  fclose(fout);

  return Ok;    
}

// use short can id for scan_table

unsigned char ShortCanId(unsigned short id)
{
  return (unsigned char)((id & 0x780) / 8 + (id & 0xf));
}

unsigned short LongCanId(unsigned char id)
{
  return (unsigned short)(8*(id & 0xf0) + (id & 0xf));
}


