/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include <Python.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "KCanElster.h"
#include "KElsterTable.h"

static KCanElster Can;
static bool CanIsInit = false;
static bool trace = false;
static char can_dev[128] = "can0";

static PyObject * elsterError;

static bool InitCan()
{
  if (CanIsInit)
    return false;
  
  if (!Can.Init(can_dev, 0x700))
  {
    char err[128];
    sprintf(err, "device \"%s\" not connected\n", can_dev);
    
    if (trace)
      printf("%s", err);
    PyErr_SetString(elsterError, err);
    return false;
  }
  CanIsInit = true;
  Can.Resume();
  
  if (trace)
    printf("can connected\n");
  
  return true;
}
  
static bool get_command(PyObject *args, const char * & command)
{
  if (!PyArg_ParseTuple(args, "s", &command))
    return false;
  
  if (!command)
    return false;
  
  while (*command == ' ')
    command++;
  
  return true;  
}

static bool get_unsigned_short(const char * & command, unsigned short & res)
{
  if (!NCanUtils::IsHexDigit(*command))
    return false;
  
  unsigned result = NCanUtils::GetHex(command);
  while (*command == ' ')
    command++;
  
  res = (unsigned short) result;
  
  return result <= 0xffff;  
}
  
static bool elsterStart(PyObject *args, unsigned short & send_id,
                        unsigned short & recv_id, unsigned short & elster_index)
{
  const char * command;

  if (!get_command(args, command))
    return false;
  
  bool Ok = get_unsigned_short(command, send_id);
  if (Ok)
    Ok = get_unsigned_short(command, recv_id);
  if (Ok)
    Ok = get_unsigned_short(command, elster_index);
  if (*command)
    Ok = false;
  
  if (!Ok)
  {
    PyErr_SetString(elsterError, "invalid parameter");
    return false;
  }
  
  if (send_id != 0x680 && send_id != 0x700 && send_id != 0x780)
    return false;
  
  if (!CanIsInit)
    if (!InitCan())
      return false;
  
  Can.Trace = trace;
  Can.SendCanId = send_id;
  
  return true;    
}
  
  
////////////////////////////////////////////////////////////////////////////////

// siehe: https://docs.python.org/3/extending/extending.html

#ifdef __cplusplus
extern "C" {
#endif

#define PY_RETURN_NONE Py_INCREF(Py_None), Py_None
  
static PyObject * elster_getname(PyObject *self, PyObject *args)
{
  const char * command;
  unsigned short elster_index;
  
  if (!get_command(args, command))
    return PY_RETURN_NONE;
  
  if (!get_unsigned_short(command, elster_index))
    return PY_RETURN_NONE;
      
  if (*command)
    return PY_RETURN_NONE;
  
  const ElsterIndex * elst_ind = GetElsterIndex(elster_index);
  if (elst_ind)
    return PyBytes_FromString(elst_ind->Name);
  
  return PyLong_FromLong(-1);
}

static PyObject * elster_gettype(PyObject *self, PyObject *args)
{
  const char * command;
  unsigned short elster_index;

  if (!get_command(args, command))
    return PY_RETURN_NONE;
  
  if (!get_unsigned_short(command, elster_index))
    return PY_RETURN_NONE;

  if (*command)
    return PY_RETURN_NONE;
  
  const ElsterIndex * elst_ind = GetElsterIndex(elster_index);
  if (elst_ind)
    return PyBytes_FromString(ElsterTypeToName(elst_ind->Type));
  
  return PyLong_FromLong(-1);
}
  
static PyObject * elster_getvalue(PyObject *self, PyObject *args)
{
  unsigned short send_id;
  unsigned short recv_id;
  unsigned short elster_index;
  unsigned short value;
  
  if (!elsterStart(args, send_id, recv_id, elster_index))
    return PY_RETURN_NONE;
  
  if (!Can.GetValue((unsigned short) recv_id, elster_index, value))
    return PY_RETURN_NONE;
 
  return PyLong_FromLong(value);
}

static PyObject * elster_getstring(PyObject *self, PyObject *args)
{
  unsigned short send_id;
  unsigned short recv_id;
  unsigned short elster_index;
  unsigned short value;
  char Val[32];
  
  if (!elsterStart(args, send_id, recv_id, elster_index))
    return elsterError;
  
  if (!Can.GetValue(recv_id, elster_index, value))
    return PY_RETURN_NONE;

  const ElsterIndex * elst_ind = GetElsterIndex(elster_index);
  if (elst_ind)
  {
    bool double_val = false;
    unsigned short value1;
    unsigned short value2;
    
    if ((value & 0xff00) != 0x8000)
    {
      if (elst_ind->Type == et_double_val ||
          elst_ind->Type == et_triple_val)
      {
        double_val = Can.GetValue(recv_id, ++elster_index, value1);
        if (double_val && elst_ind->Type == et_triple_val)
          double_val = Can.GetValue(recv_id, ++elster_index, value2);
      }
    }
    if (double_val)
    {
      const ElsterIndex * elst_ind1 = GetElsterIndex(elster_index);
      if (elst_ind1)
        elst_ind = elst_ind1;
      
      double d = (double)(value) / 1000.0 + (double)(value1);
      if (elst_ind->Type == et_triple_val)
      {
        d /= 1000.0;
        d += (double)(value2);
        sprintf(Val, "%.6f", d);
      } else
        sprintf(Val, "%.3f", d);
    } else
      SetValueType(Val, elst_ind->Type, value);
  } else
    sprintf(Val, "%d", value);
  
  return PyBytes_FromString(Val);
}
  
static PyObject * elster_trace(PyObject *self, PyObject *args)
{
  char str[64];
  
  trace = !trace;
  strcpy(str, trace ? "trace=true" : "trace=false");
  return PyBytes_FromString(str);
}

static PyObject * elster_setdev(PyObject *self, PyObject *args)
{
  const char * command;
  
  if (!get_command(args, command) ||
      strlen(command) >= High(can_dev))
    return PY_RETURN_NONE;
  
  strcpy(can_dev, command);
  return PyBytes_FromString(can_dev);
}
  
static PyObject * elster_system(PyObject *self, PyObject *args)
{
  const char *command;
  int sts;
  
  if (!PyArg_ParseTuple(args, "s", &command))
    return PY_RETURN_NONE;
  
  sts = system(command);
  if (sts < 0)
  {
    PyErr_SetString(elsterError, "System command failed");
    return PY_RETURN_NONE;
  }
  return PyLong_FromLong(sts);
}

static PyMethodDef elsterMethods[] =
{
  {"system",    elster_system,    METH_VARARGS, "Execute a shell command."},
  {"getvalue",  elster_getvalue,  METH_VARARGS, "Get Elster value"},
  {"getstring", elster_getstring, METH_VARARGS, "Get Elster string"},
  {"getname",   elster_getname,   METH_VARARGS, "Get Elster name"},
  {"gettype",   elster_gettype,   METH_VARARGS, "Get Elster type"},
  {"trace",     elster_trace,     METH_VARARGS, "Set tracing"},
  {"setdev",    elster_setdev,    METH_VARARGS, "Set can device"},
  {NULL, NULL, 0, NULL}
};

////////////////////////////////////////////////////////////////////////////////

struct module_state
{
  PyObject * error;
};

#define GETSTATE(m) ((struct module_state*)PyModule_GetState(m))
  
static int elster_traverse(PyObject *m, visitproc visit, void *arg)
{
  Py_VISIT(GETSTATE(m)->error);
  return 0;
}
  
static int elster_clear(PyObject *m)
{
  Py_CLEAR(GETSTATE(m)->error);
  
  return 0;
}

static void elster_free(void *)
{
  if (Can.SendCanId)
    Can.Halt();  
}
  
static struct PyModuleDef moduledef =
{
  PyModuleDef_HEAD_INIT,
  "elster",
  NULL,
  sizeof(struct module_state), // Wenn hier -1 steht, dann wird elster_free nicht ausgefhrt?
  elsterMethods,
  NULL,
  elster_traverse,
  elster_clear,
  elster_free
};
  
PyObject * PyInit_elster(void)
{
  PyObject * m;

  m = PyModule_Create(&moduledef);
  if (!m)
    return NULL;

  elsterError = PyErr_NewException("elster.error", NULL, NULL);
  Py_INCREF(elsterError);
  PyModule_AddObject(m, "error", elsterError);

  return m;
}

#ifdef __cplusplus
}
#endif

int main(int argc, char *argv[])
{
  Can.SendCanId = 0;
  wchar_t name[1024];
  
  for (unsigned i = 0; i <= strlen(argv[0]); i++)
    name[i] = argv[0][i];
  
  Py_SetProgramName(name);
  Py_Initialize();
  
  return 0;
}

